@@ -1,7 +1,9 @@
 # -*- coding: utf-8 -*-
 
+from __future__ import with_statement
 from codecs import EncodedFile
 from hashlib import md5
+import atexit
 import logging
 import logging.handlers
 import os
@@ -11,9 +13,11 @@ import warnings
 from time import localtime, strftime, time
 
 from meta import name as appname, script2pywname
+from multiprocess import mp
 from options import debug
 from safe_print import SafePrinter, safe_print as _safe_print
 from util_io import StringIOu as StringIO
+from util_os import safe_glob
 from util_str import safe_str, safe_unicode
 
 logging.raiseExceptions = 0
@@ -29,6 +33,7 @@ logger = None
 
 
 def showwarning(message, category, filename, lineno, file=None, line=""):
+	# Adapted from _showwarning in Python2.7/lib/logging/__init__.py
 	"""
 	Implementation of showwarnings which redirects to logging, which will first
 	check to see if the file parameter is None. If a file is specified, it will
@@ -48,11 +53,11 @@ def showwarning(message, category, filename, lineno, file=None, line=""):
 		logger = logging.getLogger("py.warnings")
 		if not logger.handlers:
 			if hasattr(sys.stderr, "isatty") and sys.stderr.isatty():
-				handler = logging.StreamHandler()
+				handler = logging.StreamHandler()  # Logs to stderr by default
 			else:
 				handler = logging.NullHandler()
 			logger.addHandler(handler)
-		logger.warning("%s", s.strip())
+		log(s.strip(), fn=logger.warning)
 
 warnings.showwarning = showwarning
 
@@ -63,8 +68,7 @@ def wx_log(logwindow, msg):
 	if logwindow.IsShownOnScreen():
 		# Check if log buffer has been emptied or not.
 		# If it has, our log message is already included.
-		logbuffer.seek(0)
-		if logbuffer.read():
+		if logbuffer.tell():
 			logwindow.Log(msg)
 
 
@@ -108,7 +112,7 @@ class Log():
 		if fn:
 			for line in msg.split("\n"):
 				fn(line)
-		if "wx" in sys.modules:
+		if "wx" in sys.modules and mp.current_process().name == "MainProcess":
 			from wxaddons import wx
 			if wx.GetApp() is not None and \
 			   hasattr(wx.GetApp(), "frame") and \
@@ -175,10 +179,10 @@ safe_print = SafeLogger()
 
 
 def get_file_logger(name, level=loglevel, when="midnight", backupCount=5,
-					logdir=None, filename=None):
+					logdir=None, filename=None, confighome=None):
 	""" Return logger object.
 	
-	A TimedRotatingFileHandler (if when == "never") or FileHandler will be used.
+	A TimedRotatingFileHandler or FileHandler (if when == "never") will be used.
 	
 	"""
 	global _logdir
@@ -187,6 +191,80 @@ def get_file_logger(name, level=loglevel, when="midnight", backupCount=5,
 	logger = logging.getLogger(name)
 	if not filename:
 		filename = name
+	mode = "a"
+	if confighome:
+		# Use different logfile name (append number) for each additional
+		# instance
+		is_main_process = mp.current_process().name == "MainProcess"
+		if os.path.basename(confighome).lower() == "dispcalgui":
+			lockbasename = filename.replace(appname, "dispcalGUI")
+		else:
+			lockbasename = filename
+		lockfilepath = os.path.join(confighome, lockbasename + ".lock")
+		if os.path.isfile(lockfilepath):
+			try:
+				with open(lockfilepath, "r") as lockfile:
+					instances = len(filter(lambda s: s.strip(),
+										   lockfile.readlines()))
+			except:
+				pass
+			else:
+				if not is_main_process:
+					# Running as child from multiprocessing under Windows
+					instances -= 1
+				if instances:
+					filenames = [filename]
+					filename += ".%i" % instances
+					filenames.append(filename)
+					if filenames[0].endswith("-apply-profiles"):
+						# Running the profile loader always sends a close
+						# request to an already running instance, so there
+						# will be at most two logfiles, and we want to use
+						# the one not currently in use.
+						mtimes = {}
+						for filename in filenames:
+							logfile = os.path.join(logdir, filename + ".log")
+							if not os.path.isfile(logfile):
+								mtimes[0] = filename
+								continue
+							try:
+								logstat = os.stat(logfile)
+							except Exception, exception:
+								safe_print(u"Warning - os.stat('%s') failed: %s" % 
+										   tuple(safe_unicode(s) for s in (logfile,
+																		   exception)))
+							else:
+								mtimes[logstat.st_mtime] = filename
+						if mtimes:
+							filename = mtimes[sorted(mtimes.keys())[0]]
+		if is_main_process:
+			for lockfilepath in safe_glob(os.path.join(confighome,
+													   lockbasename +
+													   ".mp-worker-*.lock")):
+				try:
+					os.remove(lockfilepath)
+				except:
+					pass
+		else:
+			# Running as child from multiprocessing under Windows
+			lockbasename += ".mp-worker-"
+			process_num = 1
+			while os.path.isfile(os.path.join(confighome,
+											  lockbasename + "%i.lock" % 
+											  process_num)):
+				process_num += 1
+			lockfilepath = os.path.join(confighome,
+										lockbasename + "%i.lock" % process_num)
+			try:
+				with open(lockfilepath, "w") as lockfile:
+					pass
+			except:
+				pass
+			else:
+				atexit.register(os.remove, lockfilepath)
+			when = "never"
+			filename += ".mp-worker-%i" % process_num
+			mode = "w"
 	logfile = os.path.join(logdir, filename + ".log")
 	for handler in logger.handlers:
 		if (isinstance(handler, logging.FileHandler) and
@@ -287,7 +365,7 @@ def get_file_logger(name, level=loglevel, when="midnight", backupCount=5,
 																		when=when,
 																		backupCount=backupCount)
 			else:
-				filehandler = logging.FileHandler(logfile)
+				filehandler = logging.FileHandler(logfile, mode)
 			fileformatter = logging.Formatter("%(asctime)s %(message)s")
 			filehandler.setFormatter(fileformatter)
 			logger.addHandler(filehandler)
@@ -297,7 +375,8 @@ def get_file_logger(name, level=loglevel, when="midnight", backupCount=5,
 	return logger
 
 
-def setup_logging(logdir, name=appname, ext=".py", backupCount=5):
+def setup_logging(logdir, name=appname, ext=".py", backupCount=5,
+				  confighome=None):
 	"""
 	Setup the logging facility.
 	"""
@@ -307,7 +386,8 @@ def setup_logging(logdir, name=appname, ext=".py", backupCount=5):
 	if (name.startswith(appname) or name.startswith("dispcalGUI") or
 		ext in (".app", ".exe", ".pyw")):
 		logger = get_file_logger(None, loglevel, "midnight",
-								 backupCount, filename=name)
+								 backupCount, filename=name,
+								 confighome=confighome)
 		if name == appname or name == "dispcalGUI":
 			streamhandler = logging.StreamHandler(logbuffer)
 			streamformatter = logging.Formatter("%(asctime)s %(message)s")
